@@ -30,12 +30,16 @@ namespace Python.EmbeddingTest
             /* Append the tests directory to sys.path
              * using reflection to circumvent the private
              * modifiers placed on most Runtime methods. */
+#if NETCOREAPP
+            const string s = "../../fixtures";
+#else
             const string s = "../fixtures";
+#endif
             string testPath = Path.Combine(TestContext.CurrentContext.TestDirectory, s);
 
             IntPtr str = Runtime.Runtime.PyString_FromString(testPath);
             IntPtr path = Runtime.Runtime.PySys_GetObject("path");
-            Runtime.Runtime.PyList_Append(path, str);
+            Runtime.Runtime.PyList_Append(new BorrowedReference(path), str);
         }
 
         [TearDown]
