@@ -19,6 +19,15 @@ PyNet_Args *PyNet_Init(int ext)
     pn_args->shutdown = NULL;
     pn_args->module = NULL;
 
+#ifdef __linux__
+    // Force preload libmono-2.0 as global. Without this, on some systems 
+    // symbols from libmono are not found by libmononative (which implements
+    // some of the System.* namespaces). Since the only happened on Linux so
+    // far, we hardcode the library name, load the symbols into the global
+    // namespace and leak the handle.
+    dlopen("libmono-2.0.so", RTLD_LAZY | RTLD_GLOBAL);
+#endif
+
     if (ext == 0)
     {
         pn_args->init_name = "Python.Runtime:Initialize()";
@@ -91,9 +100,9 @@ MonoMethod *getMethodFromClass(MonoClass *cls, char *name)
     return method;
 }
 
-void main_thread_handler(gpointer user_data)
+void main_thread_handler(PyNet_Args *user_data)
 {
-    PyNet_Args *pn_args = (PyNet_Args *)user_data;
+    PyNet_Args *pn_args = user_data;
     MonoMethod *init;
     MonoImage *pr_image;
     MonoClass *pythonengine;
@@ -122,7 +131,7 @@ void main_thread_handler(gpointer user_data)
             strcpy(new_ld_library_path, py_libdir);
             strcat(new_ld_library_path, ":");
             strcat(new_ld_library_path, ld_library_path);
-            setenv("LD_LIBRARY_PATH", py_libdir, 1);
+            setenv("LD_LIBRARY_PATH", new_ld_library_path, 1);
         }
     }
 
@@ -135,7 +144,7 @@ void main_thread_handler(gpointer user_data)
     int ii = 0;
     for (ii = 0; ii < PyList_Size(syspath); ++ii)
     {
-#if PY_MAJOR_VERSION > 2
+#if PY_MAJOR_VERSION >= 3
         Py_ssize_t wlen;
         wchar_t *wstr = PyUnicode_AsWideCharString(PyList_GetItem(syspath, ii), &wlen);
         char *pydir = (char*)malloc(wlen + 1);
@@ -150,7 +159,7 @@ void main_thread_handler(gpointer user_data)
         strncpy(curdir, strlen(pydir) > 0 ? pydir : ".", 1024);
         strncat(curdir, slash, 1024);
 
-#if PY_MAJOR_VERSION > 2
+#if PY_MAJOR_VERSION >= 3
         free(pydir);
 #endif
 
@@ -241,7 +250,7 @@ void main_thread_handler(gpointer user_data)
 // Get string from a Mono exception
 char *PyNet_ExceptionToString(MonoObject *e)
 {
-    MonoMethodDesc *mdesc = mono_method_desc_new(":ToString()", FALSE);
+    MonoMethodDesc *mdesc = mono_method_desc_new(":ToString()", 0 /*FALSE*/);
     MonoMethod *mmethod = mono_method_desc_search_in_class(mdesc, mono_get_object_class());
     mono_method_desc_free(mdesc);
 
