@@ -4,6 +4,7 @@ using System.Collections.Concurrent;
 using System.Collections.Generic;
 using System.Diagnostics;
 using System.IO;
+using System.Linq;
 using System.Reflection;
 using System.Threading;
 
@@ -17,16 +18,24 @@ namespace Python.Runtime
     {
         // modified from event handlers below, potentially triggered from different .NET threads
         // therefore this should be a ConcurrentDictionary
-        private static ConcurrentDictionary<string, ConcurrentDictionary<Assembly, string>> namespaces;
+        //
+        // WARNING: Dangerous if cross-app domain usage is ever supported
+        //    Reusing the dictionary with assemblies accross multiple initializations is problematic. 
+        //    Loading happens from CurrentDomain (see line 53). And if the first call is from AppDomain that is later unloaded, 
+        //    than it can end up referring to assemblies that are already unloaded (default behavior after unload appDomain - 
+        //     unless LoaderOptimization.MultiDomain is used);
+        //    So for multidomain support it is better to have the dict. recreated for each app-domain initialization
+        private static ConcurrentDictionary<string, ConcurrentDictionary<Assembly, string>> namespaces =
+            new ConcurrentDictionary<string, ConcurrentDictionary<Assembly, string>>();
         //private static Dictionary<string, Dictionary<string, string>> generics;
         private static AssemblyLoadEventHandler lhandler;
         private static ResolveEventHandler rhandler;
 
         // updated only under GIL?
-        private static Dictionary<string, int> probed;
+        private static Dictionary<string, int> probed = new Dictionary<string, int>(32);
 
         // modified from event handlers below, potentially triggered from different .NET threads
-        private static AssemblyList assemblies;
+        private static ConcurrentQueue<Assembly> assemblies;
         internal static List<string> pypath;
 
         private AssemblyManager()
@@ -40,10 +49,7 @@ namespace Python.Runtime
         /// </summary>
         internal static void Initialize()
         {
-            namespaces = new ConcurrentDictionary<string, ConcurrentDictionary<Assembly, string>>();
-            probed = new Dictionary<string, int>(32);
-            //generics = new Dictionary<string, Dictionary<string, string>>();
-            assemblies = new AssemblyList(16);
+            assemblies = new ConcurrentQueue<Assembly>();
             pypath = new List<string>(16);
 
             AppDomain domain = AppDomain.CurrentDomain;
@@ -60,7 +66,7 @@ namespace Python.Runtime
                 try
                 {
                     ScanAssembly(a);
-                    assemblies.Add(a);
+                    assemblies.Enqueue(a);
                 }
                 catch (Exception ex)
                 {
@@ -91,7 +97,7 @@ namespace Python.Runtime
         private static void AssemblyLoadHandler(object ob, AssemblyLoadEventArgs args)
         {
             Assembly assembly = args.LoadedAssembly;
-            assemblies.Add(assembly);
+            assemblies.Enqueue(assembly);
             ScanAssembly(assembly);
         }
 
@@ -132,14 +138,14 @@ namespace Python.Runtime
         internal static void UpdatePath()
         {
             IntPtr list = Runtime.PySys_GetObject("path");
-            int count = Runtime.PyList_Size(list);
+            var count = Runtime.PyList_Size(list);
             if (count != pypath.Count)
             {
                 pypath.Clear();
                 probed.Clear();
                 for (var i = 0; i < count; i++)
                 {
-                    IntPtr item = Runtime.PyList_GetItem(list, i);
+                    BorrowedReference item = Runtime.PyList_GetItem(list, i);
                     string path = Runtime.GetManagedString(item);
                     if (path != null)
                     {
@@ -340,12 +346,14 @@ namespace Python.Runtime
         /// </summary>
         internal static void ScanAssembly(Assembly assembly)
         {
+            if (assembly.GetCustomAttribute<PyExportAttribute>()?.Export == false)
+            {
+                return;
+            }
             // A couple of things we want to do here: first, we want to
             // gather a list of all of the namespaces contributed to by
             // the assembly.
-
-            Type[] types = assembly.GetTypes();
-            foreach (Type t in types)
+            foreach (Type t in GetTypes(assembly))
             {
                 string ns = t.Namespace ?? "";
                 if (!namespaces.ContainsKey(ns))
@@ -419,10 +427,9 @@ namespace Python.Runtime
             {
                 foreach (Assembly a in namespaces[nsname].Keys)
                 {
-                    Type[] types = a.GetTypes();
-                    foreach (Type t in types)
+                    foreach (Type t in GetTypes(a))
                     {
-                        if ((t.Namespace ?? "") == nsname)
+                        if ((t.Namespace ?? "") == nsname && !t.IsNested)
                         {
                             names.Add(t.Name);
                         }
@@ -449,12 +456,13 @@ namespace Python.Runtime
         /// looking in the currently loaded assemblies for the named
         /// type. Returns null if the named type cannot be found.
         /// </summary>
+        [Obsolete("Use LookupTypes and handle name conflicts")]
         public static Type LookupType(string qname)
         {
             foreach (Assembly assembly in assemblies)
             {
                 Type type = assembly.GetType(qname);
-                if (type != null)
+                if (type != null && IsExported(type))
                 {
                     return type;
                 }
@@ -463,101 +471,40 @@ namespace Python.Runtime
         }
 
         /// <summary>
-        /// Wrapper around List&lt;Assembly&gt; for thread safe access
+        /// Returns the <see cref="Type"/> objects for the given qualified name,
+        /// looking in the currently loaded assemblies for the named
+        /// type.
         /// </summary>
-        private class AssemblyList : IEnumerable<Assembly>
+        public static IEnumerable<Type> LookupTypes(string qualifiedName)
+            => assemblies.Select(assembly => assembly.GetType(qualifiedName)).Where(type => type != null && IsExported(type));
+
+        internal static Type[] GetTypes(Assembly a)
         {
-            private readonly List<Assembly> _list;
-            private readonly ReaderWriterLockSlim _lock;
-
-            public AssemblyList(int capacity)
+            if (a.IsDynamic)
             {
-                _list = new List<Assembly>(capacity);
-                _lock = new ReaderWriterLockSlim();
-            }
-
-            public int Count
-            {
-                get
-                {
-                    _lock.EnterReadLock();
-                    try
-                    {
-                        return _list.Count;
-                    }
-                    finally
-                    {
-                        _lock.ExitReadLock();
-                    }
-                }
-            }
-
-            public void Add(Assembly assembly)
-            {
-                _lock.EnterWriteLock();
                 try
                 {
-                    _list.Add(assembly);
+                    return a.GetTypes().Where(IsExported).ToArray();
                 }
-                finally
+                catch (ReflectionTypeLoadException exc)
                 {
-                    _lock.ExitWriteLock();
+                    // Return all types that were successfully loaded
+                    return exc.Types.Where(x => x != null && IsExported(x)).ToArray();
                 }
             }
-
-            public IEnumerator GetEnumerator()
+            else
             {
-                return ((IEnumerable<Assembly>)this).GetEnumerator();
-            }
-
-            /// <summary>
-            /// Enumerator wrapping around <see cref="AssemblyList._list" />'s enumerator.
-            /// Acquires and releases a read lock on <see cref="AssemblyList._lock" /> during enumeration
-            /// </summary>
-            private class Enumerator : IEnumerator<Assembly>
-            {
-                private readonly AssemblyList _assemblyList;
-
-                private readonly IEnumerator<Assembly> _listEnumerator;
-
-                public Enumerator(AssemblyList assemblyList)
+                try
                 {
-                    _assemblyList = assemblyList;
-                    _assemblyList._lock.EnterReadLock();
-                    _listEnumerator = _assemblyList._list.GetEnumerator();
+                    return a.GetExportedTypes().Where(IsExported).ToArray();
                 }
-
-                public void Dispose()
+                catch (FileNotFoundException)
                 {
-                    _listEnumerator.Dispose();
-                    _assemblyList._lock.ExitReadLock();
+                    return new Type[0];
                 }
-
-                public bool MoveNext()
-                {
-                    return _listEnumerator.MoveNext();
-                }
-
-                public void Reset()
-                {
-                    _listEnumerator.Reset();
-                }
-
-                public Assembly Current
-                {
-                    get { return _listEnumerator.Current; }
-                }
-
-                object IEnumerator.Current
-                {
-                    get { return Current; }
-                }
-            }
-
-            IEnumerator<Assembly> IEnumerable<Assembly>.GetEnumerator()
-            {
-                return new Enumerator(this);
             }
         }
+
+        static bool IsExported(Type type) => type.GetCustomAttribute<PyExportAttribute>()?.Export != false;
     }
 }
