@@ -23,6 +23,32 @@ namespace Python.Runtime
     /// </summary>
     internal class NativeCall
     {
+#if NETSTANDARD
+        [UnmanagedFunctionPointer(CallingConvention.Cdecl)]
+        private delegate void Void_1_Delegate(IntPtr a1);
+
+        [UnmanagedFunctionPointer(CallingConvention.Cdecl)]
+        private delegate int Int_3_Delegate(IntPtr a1, IntPtr a2, IntPtr a3);
+
+        public static void Void_Call_1(IntPtr fp, IntPtr a1)
+        {
+            var d = Marshal.GetDelegateForFunctionPointer<Interop.DestructorFunc>(fp);
+            d(a1);
+        }
+
+        public static IntPtr Call_3(IntPtr fp, IntPtr a1, IntPtr a2, IntPtr a3)
+        {
+            var d = Marshal.GetDelegateForFunctionPointer<Interop.TernaryFunc>(fp);
+            return d(a1, a2, a3);
+        }
+
+
+        public static int Int_Call_3(IntPtr fp, IntPtr a1, IntPtr a2, IntPtr a3)
+        {
+            var d = Marshal.GetDelegateForFunctionPointer<Interop.ObjObjArgFunc>(fp);
+            return d(a1, a2, a3);
+        }
+#else
         private static AssemblyBuilder aBuilder;
         private static ModuleBuilder mBuilder;
 
@@ -132,9 +158,10 @@ namespace Python.Runtime
         {
             return Impl.Int_Call_3(fp, a1, a2, a3);
         }
+#endif
     }
 
-
+#if !NETSTANDARD
     /// <summary>
     /// Defines native call signatures to be generated by NativeCall.
     /// </summary>
@@ -148,4 +175,5 @@ namespace Python.Runtime
 
         IntPtr Call_3(IntPtr funcPtr, IntPtr a1, IntPtr a2, IntPtr a3);
     }
+#endif
 }
