@@ -4,14 +4,15 @@ namespace Python.Runtime
 {
     /// <summary>
     /// Represents a standard Python list object. See the documentation at
-    /// http://www.python.org/doc/current/api/listObjects.html for details.
+    /// PY2: https://docs.python.org/2/c-api/list.html
+    /// PY3: https://docs.python.org/3/c-api/list.html
+    /// for details.
     /// </summary>
     public class PyList : PySequence
     {
         /// <summary>
         /// PyList Constructor
         /// </summary>
-        ///
         /// <remarks>
         /// Creates a new PyList from an existing object reference. Note
         /// that the instance assumes ownership of the object reference.
@@ -25,13 +26,12 @@ namespace Python.Runtime
         /// <summary>
         /// PyList Constructor
         /// </summary>
-        ///
         /// <remarks>
         /// Copy constructor - obtain a PyList from a generic PyObject. An
         /// ArgumentException will be thrown if the given object is not a
         /// Python list object.
         /// </remarks>
-        public PyList(PyObject o) : base()
+        public PyList(PyObject o)
         {
             if (!IsListType(o))
             {
@@ -45,11 +45,10 @@ namespace Python.Runtime
         /// <summary>
         /// PyList Constructor
         /// </summary>
-        ///
         /// <remarks>
         /// Creates a new empty Python list object.
         /// </remarks>
-        public PyList() : base()
+        public PyList()
         {
             obj = Runtime.PyList_New(0);
             if (obj == IntPtr.Zero)
@@ -62,15 +61,14 @@ namespace Python.Runtime
         /// <summary>
         /// PyList Constructor
         /// </summary>
-        ///
         /// <remarks>
         /// Creates a new Python list object from an array of PyObjects.
         /// </remarks>
-        public PyList(PyObject[] items) : base()
+        public PyList(PyObject[] items)
         {
             int count = items.Length;
             obj = Runtime.PyList_New(count);
-            for (int i = 0; i < count; i++)
+            for (var i = 0; i < count; i++)
             {
                 IntPtr ptr = items[i].obj;
                 Runtime.XIncref(ptr);
@@ -86,7 +84,6 @@ namespace Python.Runtime
         /// <summary>
         /// IsListType Method
         /// </summary>
-        ///
         /// <remarks>
         /// Returns true if the given object is a Python list.
         /// </remarks>
@@ -99,7 +96,6 @@ namespace Python.Runtime
         /// <summary>
         /// AsList Method
         /// </summary>
-        ///
         /// <remarks>
         /// Converts a Python object to a Python list if possible, raising
         /// a PythonException if the conversion is not possible. This is
@@ -119,13 +115,12 @@ namespace Python.Runtime
         /// <summary>
         /// Append Method
         /// </summary>
-        ///
         /// <remarks>
         /// Append an item to the list object.
         /// </remarks>
         public void Append(PyObject item)
         {
-            int r = Runtime.PyList_Append(obj, item.obj);
+            int r = Runtime.PyList_Append(this.Reference, item.obj);
             if (r < 0)
             {
                 throw new PythonException();
@@ -135,13 +130,12 @@ namespace Python.Runtime
         /// <summary>
         /// Insert Method
         /// </summary>
-        ///
         /// <remarks>
         /// Insert an item in the list object at the given index.
         /// </remarks>
         public void Insert(int index, PyObject item)
         {
-            int r = Runtime.PyList_Insert(obj, index, item.obj);
+            int r = Runtime.PyList_Insert(this.Reference, index, item.obj);
             if (r < 0)
             {
                 throw new PythonException();
@@ -152,13 +146,12 @@ namespace Python.Runtime
         /// <summary>
         /// Reverse Method
         /// </summary>
-        ///
         /// <remarks>
         /// Reverse the order of the list object in place.
         /// </remarks>
         public void Reverse()
         {
-            int r = Runtime.PyList_Reverse(obj);
+            int r = Runtime.PyList_Reverse(this.Reference);
             if (r < 0)
             {
                 throw new PythonException();
@@ -169,17 +162,16 @@ namespace Python.Runtime
         /// <summary>
         /// Sort Method
         /// </summary>
-        ///
         /// <remarks>
         /// Sort the list in place.
         /// </remarks>
         public void Sort()
         {
-            int r = Runtime.PyList_Sort(obj);
+            int r = Runtime.PyList_Sort(this.Reference);
             if (r < 0)
             {
                 throw new PythonException();
             }
         }
     }
-}
\ No newline at end of file
+}
