@@ -1,11 +1,8 @@
-using System;
-
 namespace Python.Test
 {
-    //========================================================================
-    // Supports CLR class unit tests.
-    //========================================================================
-
+    /// <summary>
+    /// Supports CLR class unit tests.
+    /// </summary>
     public delegate void PublicDelegate();
 
     internal delegate void InternalDelegate();
@@ -60,4 +57,4 @@ namespace Python.Test
             return d();
         }
     }
-}
\ No newline at end of file
+}
